from django.test import TestCase

from django_policies.policies import Policy
from django_policies.registry import (
    PolicyRegistry,
    autodiscover_policies,
    default_registry,
)
from testapp.models import Article


class RegistryTests(TestCase):
    def test_get_permissions_for_model(self):
        r = PolicyRegistry()
        permissions = r.get_permissions_for_model(Article)

        expected = [
            "testapp.add_article",
            "testapp.change_article",
            "testapp.delete_article",
            "testapp.view_article",
            "testapp.publish_article",
        ]

        self.assertEqual(len(permissions), len(expected))
        for e in expected:
            self.assertIn(e, permissions)

    def test_register(self):
        r = PolicyRegistry()

        class TestPolicy(Policy):
            pass

        r.register(Article, TestPolicy)
        self.assertIn("testapp.add_article", r._registry)
        self.assertEqual(r._registry["testapp.add_article"], TestPolicy)

    def test_register_twice(self):
        r = PolicyRegistry()

        class TestPolicy(Policy):
            pass

        class TestPolicy2(Policy):
            pass

        r.register(Article, TestPolicy)
        with self.assertRaises(ValueError):
            r.register(Article, TestPolicy2)
        self.assertEqual(r._registry["testapp.add_article"], TestPolicy)

    def test_unregister(self):
        r = PolicyRegistry()

        class TestPolicy(Policy):
            pass

        r.register(Article, TestPolicy)
        r.unregister(Article)
        self.assertEqual(r._registry, {})

    def test_unregister_unknown(self):
        r = PolicyRegistry()
        r.unregister(Article)
        self.assertEqual(r._registry, {})

    def test_autodiscover(self):
        autodiscover_policies()

        self.assertIn("testapp.add_article", default_registry._registry)

        from testapp.policies import ArticlePolicy

        self.assertEqual(
            default_registry._registry["testapp.add_article"], ArticlePolicy
        )
        default_registry.unregister(Article)



from django.db import models
from django.conf import settings


class Article(models.Model):
    name = models.CharField(max_length=128)
    author = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)

    class Meta:
        permissions = [("publish_article", "Can publish article")]



from django_policies.policies import Policy
from django_policies.registry import default_registry


def register(model, registry=None):
    if registry is None:
        registry = default_registry

    def _policy_wrapper(policy):
        if not issubclass(policy, Policy):
            raise ValueError("Wrapped class must be a subclass of Policy.")

        registry.register(model, policy)
        return policy

    return _policy_wrapper



from django_policies.decorators import register
from django_policies.policies import Policy
from testapp.models import Article


@register(Article)
class ArticlePolicy(Policy):
    can_add_some = True

    def can_change(self, user, obj):
        return True



from typing import TYPE_CHECKING, Any
from django.apps import apps
from itertools import chain
from importlib import import_module
from django.db.models import Model

from django_policies.policies import Policy

if TYPE_CHECKING:  # pragma: no cover
    from django_policies.policies import AnyUser


class PolicyRegistry:
    _registry: dict[str, type[Policy]]

    def __init__(self) -> None:
        self._registry = {}

    def get_permissions_for_model(self, model: type[Model]) -> list[str]:
        app_label = model._meta.app_label
        model_name = model._meta.model_name
        if model_name is None:
            raise ValueError(f"model_name of {model.__class__.__name__} is None.")

        permissions = [f"{app_label}.{perm[0]}" for perm in model._meta.permissions]
        default_permissions = [
            f"{app_label}.{perm}_{model_name}"
            for perm in model._meta.default_permissions
        ]

        return list(chain(permissions, default_permissions))

    def register(self, model: type[Model], policy: type[Policy]):
        for perm in self.get_permissions_for_model(model):
            if perm in self._registry:
                raise ValueError(f"Permission {perm} is already registered.")
            self._registry[perm] = policy

    def unregister(self, model: type[Model]):
        for perm in self.get_permissions_for_model(model):
            if perm not in self._registry:
                continue
            del self._registry[perm]

    def check_permission(
        self, user: "AnyUser", permission: str, obj: Any = None
    ) -> bool:
        if permission not in self._registry:
            return False

        return self._registry[permission]().check_permission(user, permission, obj)


default_registry = PolicyRegistry()


def autodiscover_policies(module_name="policies"):
    for app in apps.get_app_configs():
        try:
            import_module(f"{app.name}.{module_name}")
        except ImportError:
            continue



from asgiref.sync import sync_to_async
from django.contrib.auth.backends import BaseBackend
from django_policies.registry import default_registry


class PolicyBackend(BaseBackend):
    async def ahas_perm(self, user_obj, perm, obj=None):
        return await sync_to_async(self.has_perm)(user_obj, perm, obj)

    def has_perm(self, user_obj, perm, obj=None) -> bool:
        return default_registry.check_permission(user_obj, perm, obj)  # pyright: ignore



from django.contrib.auth.mixins import (
    PermissionRequiredMixin as BasePermissionRequiredMixin,
)


class PermissionRequiredMixin(BasePermissionRequiredMixin):
    def get_permission_object(self):
        return None

    def has_permission(self) -> bool:
        obj = self.get_permission_object()
        perms = self.get_permission_required()
        return self.request.user.has_perms(perms, obj)



# Generated by Django 5.2.3 on 2025-06-29 22:34

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Article",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=128)),
                (
                    "author",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "permissions": [("publish_article", "Can publish article")],
            },
        ),
    ]
